#drinc:libraries/dos.g

/*
 * uudecode.drc
 *   Decode an ASCII representation into a binary file. This is the
 *   reverse of uuencode. If a parameter is given, it is the name of the
 *   uuencoded file to decode. If no parameter is given, standard input
 *   is read. The input file contains the name to use for the binary file
 *   created. It can contain more than one file, even though uuencode
 *   can't directly produce such a beast.
 *
 *   This is a very early program, written before the Amiga version of Draco
 *   had an I/O library and include files. Hence this source file is completely
 *   self contained (except for the interface routines and assembler support).
 *
 *   Program has since be updated to try to be as flexible as possible in
 *   the style of uuencoded file it can accept.
 */

uint
    BUFF_SIZE = 10000,		/* nice and big, but divisible by 3 */
    MAX_BYTES = 45,		/* maximum bytes per encoded line */
    MAX_LINE = 100;		/* maximum input line we use */

Handle_t StdOut, InFd, OutFd;

[BUFF_SIZE] char InBuff;	/* buffer for input ASCII file */
[BUFF_SIZE] byte OutBuff;	/* buffer for output binary file */
[MAX_LINE] char LineBuff;	/* buffer for input line */
uint InMax, InPos, OutPos, LineMax, LinePos;  /* buffer positions and counts */
ulong LineCount;
bool Eof;			/* true => end-of-file on input */
bool GotFile;			/* true => reading from a file */
bool IgnoreCheckSum;		/* true => ignore checksum errors */

/*
 * abort -
 *   Local routine to cleanly abort with the given error level.
 */

proc abort(uint level)void:

    if GotFile then
	Close(InFd);
    fi;
    Exit(level);
corp;

/*
 * writeString -
 *   Write a string to the standard output.
 */

proc writeString(*char s)void:
    register *char p;

    p := s;
    while p* ~= '\e' do
	p := p + sizeof(char);
    od;
    ignore Write(StdOut, s, (p - s) / sizeof(char));
corp;

/*
 * writeNumber -
 *   Write the ascii form of an integer to the standard output.
 */

proc writeNumber(register ulong n)void:
    [11] char buff;
    register *char p;

    p := &buff[5];
    p* := '\e';
    while
	p := p - sizeof(char);
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    writeString(p);
corp;

/*
 * errorN - an error message with a number in it.
 */

proc errorN(*char s1; uint n)void:

    writeString("Line ");
    writeNumber(LineCount);
    writeString(": ");
    writeString(s1);
    writeNumber(n);
    writeString("\n");
    Close(OutFd);
    abort(20);
corp;

/*
 * flush -
 *   Flush the contents of the output buffer to the output file.
 */

proc flush()void:

    if OutPos ~= 0 then
	if Write(OutFd, &OutBuff[0], OutPos) ~= OutPos then
	    writeString("Error on write - aborting.\n");
	    Close(OutFd);
	    abort(30);
	fi;
	OutPos := 0;
    fi;
corp;

/*
 * addByte -
 *   Add a single byte to the output file, flushing if needed.
 */

proc addByte(byte b)void:

    OutBuff[OutPos] := b;
    OutPos := OutPos + 1;
    if OutPos = BUFF_SIZE then
	flush();
    fi;
corp;

/*
 * getChar -
 *   Get the next character from the input binary file.
 */

proc getChar()char:

    if InPos = InMax then
	InPos := 0;
	InMax := Read(InFd, &InBuff[0], BUFF_SIZE);
	if InMax = 0 then
	    Eof := true;
	fi;
    fi;
    InPos := InPos + 1;
    InBuff[InPos - 1]
corp;

/*
 * getLine -
 *   Read an input line into LineBuff/LineMax.
 */

proc getLine()void:
    char ch;

    LineCount := LineCount + 1;
    LineMax := 0;
    while
	ch := getChar();
	not Eof and ch ~= '\n'
    do
	if LineMax < MAX_LINE - 8 then
	    if ch = '\t' then
		while
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		    LineMax % 8 ~= 0
		do
		od;
	    elif ch ~= '\r' then
		/* ignore any carriage returns that may be present */
		LineBuff[LineMax] := ch;
		LineMax := LineMax + 1;
	    fi;
	fi;
    od;
    LinePos := 0;
corp;

/*
 * getWord -
 *   Peel an input word off of the input line.
 */

proc getWord()*char:
    *char p;

    p := &LineBuff[LinePos];
    while LinePos < LineMax and LineBuff[LinePos] ~= ' ' do
	LinePos := LinePos + 1;
    od;
    if LinePos = LineMax then
	LineBuff[LinePos] := '\e';
    else
	LinePos := LinePos + 1;
    fi;
    p
corp;

/*
 * isWord -
 *   Compare two words for equality.
 */

proc isWord(register *char p, q)bool:

    while p* = q* and q* ~= '\e' do
	p := p + sizeof(char);
	q := q + sizeof(char);
    od;
    (p* = ' ' or p* = '\e') and q* = '\e'
corp;

/*
 * decodeChar -
 *   Decode a single character to a six bit value.
 */

proc decodeChar(char ch)byte:

    if ch >= ' ' and ch <= ' ' + ((1 << 6) - 1) then
	ch - ' '
    elif ch = '`' then
	0
    else
	errorN("bad encoded character: ", ch - '\e');
	0
    fi
corp;

/*
 * decodeFile -
 *   Decode the file given already opened descriptors, etc.
 */

proc decodeFile()bool:
    [MAX_BYTES] byte buff;
    register *char size;
    ulong sizeGot, sizeSaid;
    register uint count, i, j;
    register byte b1, b2, b3, b4;
    uint checksum;
    bool first, gotCheck;

    /* fetch and decode lines until we find one starting with "end " */
    first := true;
    sizeGot := 0;
    while
	getLine();
	not Eof and not isWord(getWord(), "end")
    do
	/* check for a valid length line */
	if LineMax < 1 or LineMax > MAX_BYTES * 4 / 3 + 2 then
	    errorN("invalid line length: ", LineMax);
	fi;
	/* if first line, choose checksum or not and pad the line */
	if first then
	    first := false;
	    if (LineMax - 2) % 4 = 0 then
		gotCheck := true;
	    elif (LineMax - 1) % 4 = 0 then
		gotCheck := false;
	    else
		gotCheck := false;
		while (LineMax - 1) % 4 ~= 0 do
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		od;
	    fi;
	else
	    if gotCheck then
		while (LineMax - 2) % 4 ~= 0 do
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		od;
	    else
		while (LineMax - 1) % 4 ~= 0 do
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		od;
	    fi;
	fi;
	/* get and check the actual count of encoded bytes on this line */
	count := decodeChar(LineBuff[0]);
	i := (count + 2) / 3 * 4;
	if
	    if gotCheck then
		LineMax ~= i + 2
	    else
		LineMax ~= i + 1
	    fi
	then
	    writeString("Line ");
	    writeNumber(LineCount);
	    writeString(": invalid record length: ");
	    writeNumber(count);
	    writeString(" : ");
	    writeNumber(LineMax);
	    writeString(" - padding with spaces\n");
	    if gotCheck then
		while LineMax ~= i + 2 do
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		od;
	    else
		while LineMax ~= i + 1 do
		    LineBuff[LineMax] := ' ';
		    LineMax := LineMax + 1;
		od;
	    fi;
	fi;
	/* decode the line - groups of 4 characters yield three bytes */
	if gotCheck then
	    LineMax := LineMax - 1;
	fi;
	j := 0;
	i := 1;
	checksum := 0;
	while i ~= LineMax do
	    b1 := decodeChar(LineBuff[i + 0]);
	    b2 := decodeChar(LineBuff[i + 1]);
	    b3 := decodeChar(LineBuff[i + 2]);
	    b4 := decodeChar(LineBuff[i + 3]);
	    buff[j + 0] := b1 << 2 | b2 >> 4;
	    buff[j + 1] := b2 << 4 | b3 >> 2;
	    buff[j + 2] := b3 << 6 | b4;
	    checksum := checksum + buff[j + 0] + buff[j + 1] + buff[j + 2];
	    j := j + 3;
	    i := i + 4;
	od;
	checksum := checksum % 64;
	if gotCheck and not IgnoreCheckSum and
	    checksum ~= decodeChar(LineBuff[i])
	then
	    errorN("bad checksum: ", checksum);
	fi;
	/* write out the correct number (1 - 45) of bytes */
	i := 0;
	while i ~= count do
	    addByte(buff[i]);
	    i := i + 1;
	od;
	sizeGot := sizeGot + count;
    od;
    if Eof then
	false
    else
	getLine();
	if isWord(getWord(), "size") then
	    size := getWord();
	    sizeSaid := 0;
	    while size* >= '0' and size* <= '9' do
		sizeSaid := sizeSaid * 10 + (size* - '0');
		size := size + sizeof(char);
	    od;
	    if sizeSaid ~= sizeGot then
		errorN("incorrect file size: ", sizeGot);
	    fi;
	    false
	else
	    true
	fi
    fi
corp;

/*
 * decode -
 *   Decode all of the encoded files in the already opened input file.
 */

proc decode()void:
    register *char outFileName;
    register bool gotBegin, doneFile, lineRead;

    doneFile := false;
    /* decode files until there are no more left */
    lineRead := false;
    while not Eof do
	gotBegin := false;
	/* skip lines until we find a 'begin' line or we hit end-of-file */
	while not Eof and not gotBegin do
	    if not lineRead then
		getLine();
	    fi;
	    outFileName := getWord();
	    if isWord(outFileName, "begin") then
		outFileName := getWord();
		while outFileName* >= '0' and outFileName* <= '7' do
		    outFileName := outFileName + sizeof(char);
		od;
		if outFileName* = ' ' then
		    outFileName := getWord();
		    if outFileName* ~= '\e' then
			gotBegin := true;
		    fi;
		fi;
	    fi;
	    lineRead := false;
	od;
	/* if we just found a 'begin' line, go decode an encoded file */
	if gotBegin then
	    writeString("Decoding file '");
	    writeString(outFileName);
	    writeString("'\n");
	    /* open output file, etc. */
	    DeleteFile(outFileName);
	    OutFd := Open(outFileName, MODE_NEWFILE);
	    if OutFd = 0 then
		writeString("Can't open/create output file '");
		writeString(outFileName);
		writeString("'\n");
		abort(20);
	    fi;
	    /* initialize */
	    OutPos := 0;
	    /* go decode the file */
	    lineRead := decodeFile();
	    /* flush output buffer, and close file */
	    flush();
	    Close(OutFd);
	    doneFile := true;
	fi;
    od;
    /* if we found no 'begin' line in the whole input file, then complain */
    if not doneFile then
	writeString("No 'begin' line found.\n");
	abort(20);
    fi;
corp;

/*
 * main -
 *   The main program. Get the command line parameters and parse a file
 *   name from them, else just use the standard input. The input file is
 *   decoded into 1 or more output binary files.
 */

proc main()void:
    extern GetPars(*ulong pLen; **char pPtr)void;
    *char parPtr, inFileName;
    ulong parLen;

    if OpenDosLibrary(0) ~= nil then
	IgnoreCheckSum := false;
	StdOut := Output();
	GetPars(&parLen, &parPtr);
	while parLen ~= 0 and
	    (parPtr* = ' ' or parPtr* = '\n' or parPtr* = '\r')
	do
	    parPtr := parPtr + sizeof(char);
	    parLen := parLen - 1;
	od;
	if parLen ~= 0 and parPtr* = '-' then
	    parPtr := parPtr + sizeof(char);
	    parLen := parLen - 1;
	    while parLen ~= 0 and parPtr* ~= ' ' and
		parPtr* ~= '\n' and parPtr* ~= '\r'
	    do
		case parPtr*
		incase 'c':
		incase 'C':
		    IgnoreCheckSum := true;
		default:
		    writeString("Invalid flag: '");
		    (parPtr + sizeof(char))* := '\e';
		    writeString(parPtr);
		    writeString("' - aborting\n");
		    Exit(5);
		esac;
		parPtr := parPtr + sizeof(char);
		parLen := parLen - 1;
	    od;
	    while parLen ~= 0 and parPtr* = ' ' do
		parPtr := parPtr + sizeof(char);
		parLen := parLen - 1;
	    od;
	fi;
	if parLen = 0 then
	    /* no parameter given - read from standard input */
	    InFd := Input();
	    GotFile := false;
	else
	    inFileName := parPtr;
	    while parLen ~= 0 and parPtr* ~= ' ' and
		parPtr* ~= '\n' and parPtr* ~= '\r'
	    do
		parPtr := parPtr + sizeof(char);
		parLen := parLen - 1;
	    od;
	    parPtr* := '\e';
	    /* try to open the specified input file */
	    InFd := Open(inFileName, MODE_OLDFILE);
	    if InFd = 0 then
		writeString("Can't open input file '");
		writeString(inFileName);
		writeString("'\n");
		Exit(10);
	    fi;
	    GotFile := true;
	fi;
	/* initialize for reading the input line */
	Eof := false;
	InMax := 0;
	InPos := 0;
	LineCount := 0;
	decode();
	if GotFile then
	    Close(InFd);
	fi;
	CloseDosLibrary();
    fi;
corp;
